/*
 * 北太天元（Baltamatica）示例文件。
 *
 * 本文件示例一个北太天元插件开发方法，其功能是
 * 定义 Point 类型对象以及相关运算。主要是为了演示如何进行运算符重载。
 *
 * 实现的函数：
 * 
 *     创建一个点           P = create_point(x, y)
 *     显示点的坐标         show_point(V)
 *     加法运算             z = x + y
 *     单目加               +x
 *
 * 适用于当前版本的北太天元。更多例子随软件内容更新。
 *
 * 版权所有 © 2021-2023，北太振寰（重庆）科技有限公司
 */


#include <iostream>
#include <sstream>
#include <vector>
#include "bex/bex.hpp"

#define PLUGIN_NAME "point"

using namespace baltam;

/**
 * @brief main.cpp 示例一个北太天元插件的开发方法，其功能是
 * 演示如何重载北太天元中的算子。
 *
 * 实现的函数：
 * p = create_point(x, y);
 * show_point(p);
 * p3 = add_point(p1, p2);
 * p2 = uplus_point(p1);
 *
 * 在这里，特意将单目加（uplus）运算重载为所有分量都增加 0.01
 */
void create_point(int, bxArray*[], int, const bxArray*[]);
void show_point(int, bxArray*[], int, const bxArray*[]);
void add_point(int, bxArray*[], int, const bxArray*[]);
void uplus_point(int, bxArray*[], int, const bxArray*[]);

/**
 * @brief 各个函数的帮助文字（可选）
 */
static const char * create_help = R"(
create_point 创建一个 point 类型对象

    p = create_point(x, y) 创建一个 point 对象，坐标分别为 x 和 y。
    要求 x 和 y 必须为 double 类型。
)";

static const char * show_help = R"(
show_point 在命令行中显示一个 point 的简略信息。

    show_point(p) 在命令行中 (x, y) 的方式来显示 p。

)";

/**
 * @brief Point 自定义类。
 *
 * 作为一种自定义数据类型，Point 必须继承自 batlam::extern_obj_base 类。
 * 并定义 ID 这一静态变量（会被内核使用）。
 */

struct Point : public extern_obj_base {
    double x_;
    double y_;
    BALTAM_LOCAL static int ID;

    extern_obj_base *dup() const override;
    Point();
    Point(double, double);
    Point(const Point&) = default;
    ~Point() override;
    // 可选：转化为字符串
    std::string to_string() const override;
    // 可选：类型名字
    std::string classname() const override;

    static Point * from_double(const bxArray *p);
};

int Point::ID = 0;

Point::Point() : x_(0), y_(0) {}
Point::Point(double x, double y) : x_(x), y_(y) {}

extern_obj_base *Point::dup() const {
    return new Point(*this);
}

std::string Point::to_string() const {
    std::stringstream ss;
    ss << "    (" << x_ << ", " << y_ << ")\n";
    return ss.str();
}

std::string Point::classname() const {
    return "Point";
}

Point * Point::from_double(const bxArray * p){
    if (bxIsRealDouble(p)){
        double v = *bxGetDoubles(p);
        return new Point(v, v);
    }

    return nullptr;
}

Point::~Point() {}
/**
 * @brief bxPluginInit 的定义（可选）
 *
 * bxPluginInit(args...) 会在调用 load_plugin(name, args...) 时调用，目的是
 * 进行一些初始化工作。
 *
 * 这里的初始化工作是向内核添加 Point 这个类
 * 并重载 + (bxADD_OP) 算子
 */

int bxPluginInit(int, const bxArray*[]){
    int sid = bxAddCXXClass<Point>(PLUGIN_NAME);
    // 注册加法
    bxRegisterBinaryOperator(PLUGIN_NAME, "+", sid, sid, add_point);
    bxRegisterUnaryOperator(PLUGIN_NAME, "u+", sid, uplus_point);
    // BEX v3.2 也可以用下面这个函数
    // bxRegisterOperator(PLUGIN_NAME, "+", sid, add_point);

    // 重载 disp 函数
    bxOverloadFunction(PLUGIN_NAME, "disp", show_point);
    return 0;
}

/**
 * @brief bxPluginFini 的定义（可选）
 *
 * bxPluginFini() 会在调用 unload_plugin() 时调用。目的是进行一些析构的工作。
 *
 * 这里因为无需特别的析构，所以定义为空函数。
 */

int bxPluginFini(){ return 0; }

/**
 * @brief bxPluginFunctions 的定义（必要）
 *
 * bxPluginFunctions 返回一个 bexfun_info_t * 的指针，指向函数列表的第一个元素。
 * 列表中最后一个元素的 name 标签必须为空字符串常量。
 * 函数列表可以以静态变量的方式定义在 .cpp 或 .c 中。
 *
 */
static bexfun_info_t flist[] = {
    {"create_point", create_point, create_help},
    {"show_point", show_point, show_help},
    {"", nullptr, nullptr},
};

bexfun_info_t * bxPluginFunctions(){
    return flist;
}

/**
 * @brief 以下是各种函数的实际定义。
 */

void create_point(int nlhs, bxArray *plhs[], int nrhs, const bxArray *prhs[]){
    if (nlhs > 1 && nrhs != 2){
        bxErrMsgTxt("用法：P = create_point(x, y)");
        return;
    }

    if (!bxIsDouble(prhs[0]) || !bxIsDouble(prhs[1])){
        bxErrMsgTxt("参数 1 和 2 必须为 double 类型的标量。");
        return;
    }

    Point * ret = bxNewCXXObject<Point>(
            *bxGetDoubles(prhs[0]),
            *bxGetDoubles(prhs[1]));
    plhs[0] = bxCreateExtObj(ret);
}

void show_point(int nlhs, bxArray *[], int nrhs, const bxArray *prhs[]){
    if (nlhs > 0 || nrhs != 1){
        bxReturnOverloadFailure();
        return;
    }

    Point *bv = bxGetExtObj<Point>(prhs[0]);

    if (bv == nullptr){
        bxReturnOverloadFailure();
        return;
    }

    bxPrintf(bv->to_string().c_str());
}


void add_point(int nlhs, bxArray *plhs[], int nrhs, const bxArray *prhs[]){
    if (nlhs > 1 && nrhs != 2){
        bxErrMsgTxt("用法：z = add_point(x, y)");
        return;
    }
    
    // 支持 point + point, point + double, double + point
    Point *p1 = bxGetExtObj<Point>(prhs[0]);
    Point *p2 = bxGetExtObj<Point>(prhs[1]);

    bool free_p1 = false, free_p2 = false;
    if (p1 == nullptr){
        p1 = Point::from_double(prhs[0]);
        free_p1 = true;
    }

    if (p2 == nullptr){
        p2 = Point::from_double(prhs[1]);
        free_p2 = true;
    }

    if (p1 == nullptr || p2 == nullptr){
        bxErrMsgTxt("参数 1 和 2 必须为 Point 类型对象或 double。");
        return;
    }

    Point *p3 = bxNewCXXObject<Point>(p1->x_ + p2->x_, p1->y_ + p2->y_);
    plhs[0] = bxCreateExtObj(p3);

    if (free_p1) delete p1;
    if (free_p2) delete p2;
}

void uplus_point(int nlhs, bxArray *plhs[], int nrhs, const bxArray *prhs[]){
    if (nlhs > 1 && nrhs != 1){
        bxErrMsgTxt("用法：y = uplus_point(x)");
        return;
    }
    
    Point *p1 = bxGetExtObj<Point>(prhs[0]);
    Point *p3 = bxNewCXXObject<Point>(p1->x_ + 0.01, p1->y_ + 0.01);
    plhs[0] = bxCreateExtObj(p3);
}
